from __future__ import annotations

from .exceptions import ExpressionError


class Expression:
    """
    A class for converting Python's logical operators to SQL-compatible strings
    """
    def __init__(self, query: str, table: str) -> None:
        """
        An object that represents an SQL expression/filter ("price BETWEEN 5 AND 15")

        :param query: str, e.g., id = 3571
        :param table: str, table-name
        """
        if not isinstance(query, str):
            raise TypeError(f'parameter query must be of type str, not: {type(query)}')

        self.query = query
        self.table = table

    def _validate_other(self, other) -> None:
        """
        Validate other expression is valid (for __and__() and __or__())

        :raise: ExpressionError if param: expression isn't an instance of Expression
        :raise: ExpressionError if self.table != other.table
        """
        if not isinstance(other, Expression):
            raise ExpressionError('expression must be an instance of Expression, try using a column object instead')

        if not self.table == other.table:
            raise ExpressionError(
                f'Cannot concatenate two expressions from different tables ({self.table} and {other.table})')

    def __and__(self, expression: Expression) -> Expression:
        """
        Return an Expression object with 'self.query' as: 'Expression1 AND Expression2'

        :param expression: Expression
        :return: Expression
        :raise: ExpressionError if param: expression isn't an instance of Expression
        :raise: ExpressionError if self.table != other.table
        """
        self._validate_other(other=expression)
        return Expression(query=f'{self.query} AND {expression.query}', table=self.table)

    def __or__(self, expression: Expression) -> Expression:
        """
        Return an Expression object with 'self.query' as: 'Expression1 OR Expression2'

        :param expression: Expression
        :return: Expression
        :raise: ExpressionError if param: expression isn't an instance of Expression
        :raise: ExpressionError if self.table != other.table
        """
        self._validate_other(other=expression)
        return Expression(query=f'{self.query} OR {expression.query}', table=self.table)

    def __str__(self) -> str:
        """ Get string representation of Expression instance """
        return f'SELECT ... WHERE {self.query}'

    def __repr__(self) -> str:
        """ Get string representation of Expression instance """
        return __class__.__name__ + f'(query={repr(self.query)}, table={repr(self.table)})'



from __future__ import annotations

from pandas import DataFrame

import sqlite3
from typing import Generator, Callable, Any, overload, Literal

from .exceptions import InvalidColumnError
from .column import Column
from .cache import Cache
from .expression import Expression
from .utils import create_temp_view, get_random_name, sql_tuple

PrimitiveTypes = str | int | float | bool | None
TableRow = tuple[PrimitiveTypes, ...]
ROWID = '_rowid_ AS _rowid_'  # select rowid/oid/_rowid_ as _rowid_


class IndexLoc:
    def __init__(self, table: Table) -> None:
        self.table = table
        self.len = len(table)  # to avoid recomputing

    def index_abs(self, idx: int) -> int:
        """
        Return the absolute of an index

        if the given index is negative it will convert it to positive, for ex:
        if the given index is -1 and the length of the table is 371 the method will return 371

        :param idx: int
        :return: int
        """
        if idx < 0:
            return self.len + idx
        return idx

    def validate_index(self, idx: int) -> None:
        """
        Assert given index is above zero, and below or equal to length,
        else: raise IndexError

        :param idx: int, must be positive
        :raise IndexError: if not 0 <= index < length
        :return: None
        """
        if not 0 <= idx < self.len:
            raise IndexError(f'Given index out of range ({idx})')

    @overload
    def __getitem__(self, index: int) -> TableRow:
        ...

    @overload
    def __getitem__(self, index: slice | list) -> list[TableRow]:
        ...

    def __getitem__(self, index: int | slice | list) -> TableRow | list[TableRow]:
        """
        Get row/value at given index

        if an int is passed; then values at the given index will be returned
        if a slice or a list is passed; then it will return a list of elements/cells from the table

        Note that the index is always increased by one before being passed to the SQL query,
        as the rowid columns starts the count at 1 and not 0.

        :param index: int | slice | list
        :return: tuple | list | BaseTypes
        """
        def select_cols_str() -> str:
            return ', '.join(self.table.columns)

        if isinstance(index, int):
            index = self.index_abs(index)
            self.validate_index(index)
            index += 1

            with self.table.conn as cursor:
                query = f'SELECT {select_cols_str()} FROM {self.table.name} WHERE _rowid_ == {index}'
                return cursor.execute(query).fetchall()[0]

        if isinstance(index, slice):
            indices = index.indices(self.len)
            indexes = [idx + 1 for idx in range(*indices)]

            with self.table.conn as cursor:
                query = f'SELECT {select_cols_str()} FROM {self.table.name} WHERE _rowid_ IN {sql_tuple(indexes)}'
                return cursor.execute(query).fetchall()

        if isinstance(index, list):
            indexes = [self.index_abs(idx) for idx in index]
            for idx in indexes:
                self.validate_index(idx)
            indexes = [idx + 1 for idx in indexes]
            unique_indexes = set(indexes)

            with self.table.conn as cursor:
                rows = cursor.execute(
                    f'SELECT _rowid_, {select_cols_str()} FROM {self.table.name} '
                    f'WHERE _rowid_ IN {sql_tuple(unique_indexes)}')

            idx_row_mapping: dict[int, TableRow] = {tup[0]: tup[1:] for tup in rows}  # first item in tuple is _rowid_
            return [idx_row_mapping[idx] for idx in indexes]

        raise TypeError(f'Index must be of type: int, list, or slice. not: {type(index)}')


class Table:
    """
    An object that represents an SQL table
    """
    def __init__(self, conn: sqlite3.Connection, cache: Cache, name: str) -> None:
        """
        Initialize the Table object

        :param conn: sqlite3.Connection
        :param cache: Cache, instance of Cache
        :param name: str, table name
        """
        self.conn = conn
        self.name = name
        self._cache = cache
        self._column_items: dict[str, Column] = {}

        for column in self.columns:
            self._set_column(column)

    @property
    def query(self) -> str:
        """
        Get the "SELECT *" query for the table

        The returned query is a property so if any column is added/ removed, the query will be updated.
        note that the rowid column is filtered out in 'self.columns', you can read more about that
        in the class docstring.
        """
        return f'SELECT {self._cols_as_str} FROM {self.name}'

    @property
    def columns(self) -> list[str]:
        """
        Get list with column names
        """
        with self.conn as cursor:
            return [x[1] for x in cursor.execute(f"PRAGMA table_info('{self.name}')")]

    @property
    def _cols_as_str(self) -> str:
        """
        Get a string of columns that can be passed directly to an SQL query
        """
        return ', '.join(self.columns)

    @property
    def len(self) -> int:
        """
        Return amount of rows in the table
        """
        return self._cache.execute(f'SELECT COUNT(*) FROM {self.name}')[0][0]

    @property
    def shape(self) -> tuple:
        """
        Get a tuple with: (n_rows, n_cols)
        """
        return self.len, len(self.columns)

    def describe(self) -> dict[str, dict[str, Any]]:
        """
        Get a nested dictionary with the descriptive properties for each column in the table

        :return: dict, {col1: {'min': 32, 'max': 83 ...}, col2: {'min': 'Alex', 'max': 'Zoey' ...} ...}
        """
        return {name: col.describe() for name, col in self.items()}

    def to_df(self) -> DataFrame:
        """
        Return table as a Pandas DataFrame
        """
        return DataFrame(data=iter(self), columns=self.columns)

    def data(self, limit: int = None) -> list[TableRow]:
        """
        Get table data in a nested list, ex: [('AMD', 78.54, True), ('AAPL', 125.34, True)...]

        :param limit: int
        :return: list
        """
        with self.conn as cursor:
            if limit:
                return cursor.execute(self.query + f' LIMIT {limit}').fetchall()
            return cursor.execute(self.query).fetchall()

    def sample(self, n: int = 10) -> list[TableRow]:
        """
        Get a list of random rows from the table

        :param n: int, number of rows
        :return: list with nested tuples
        """
        with self.conn as cursor:
            return cursor.execute(f'{self.query} ORDER BY RANDOM() LIMIT {n}').fetchall()

    def items(self) -> Generator[tuple[str, Column], None, None]:
        """
        Generator that yields: (column_name, col_object)
        """
        yield from self._column_items.items()

    def applymap(self, func: Callable, *, ignore_na: bool = True,
                 args: tuple = tuple(), **kwargs) -> Generator[tuple, None, None]:
        """
        Apply function on each cell in the table
        
        example:
        db = DataBase(db_path='data/forestation.db')
        table = db.regions.applymap(lambda x: len(x) if isinstance(x, str) else None)
        for row in table:
            print(row)

        (11, 3, 26, 10)
        (6, 3, 18, 19)
        (8, 3, 18, 10)
        (5, 3, 5, None)

        :param func: Callable
        :param ignore_na: bool, default: True
        :param args: tuple, args to pass to the function
        :param kwargs: keyword args to pass to the callable
        :return: Generator
        """
        for row in self:
            yield tuple(cell if cell is None and ignore_na is True else func(cell, *args, **kwargs) for cell in row)

    @property
    def iloc(self) -> IndexLoc:
        """
        Get data by: index, list, or slice

        Getitem supports three ways of indexing table rows:
        1) Singular Integer, ex: iloc[0], iloc[32], or with negative: iloc[-12]
        2) Passing a list of integers, ex: iloc[[1, 22, 4, 3, 17, 38]], iloc[[1, -4, 17, 22, 38, -4, -1]]
        4) Passing Slice, ex: iloc[:10], iloc[2:8], iloc[2:24:2]

        The return type will be a list for multiple items and a tuple for single items

        :return: tuple or list of tuples
        """
        return IndexLoc(self)

    def filter(self, expression: Expression) -> TableView:
        """
        Filter and return a table

        The way this works is you have to pass an expression, which you could either type manually:
        `from pandasdb import Expression`
        `df = db.table`
        `df.filter(Expression('col BETWEEN 3.2 AND 7.8', table_name))`

        Or by performing a logical operation with a column (just like in Pandas and Numpy)
        `df = db.table`
        `df.filter((df.col > 3.2) & (df.col < 7.8))`
        Or with `Column.between()`:
        `df.filter(df.col.between(3.2, 7.8))`

        And there is a shortcut for .filter() using square brackets:
        `df[df.col.between(3.2, 7.8)]`

        - How it works:
        Whenever you pass an expression, i.e.: `df.col == 'Jack'` a SQL query will be generated with the following:
        `SELECT * FROM table WHERE col == 'Jack'` and it will create an SQL-view to store the filtered data,
        the view itself can be filtered again and again.

        :param expression: Expression
        :return: TableView
        """
        view_name = f'_table_{self.name}_{get_random_name(size=10)}_'
        query = f"""
        SELECT ROW_NUMBER() OVER (ORDER BY _rowid_) AS _rowid_, {self._cols_as_str}
        FROM {self.name} 
        WHERE {expression.query}
        """
        return self._create_and_get_temp_view(view_name=view_name, query=query)

    def sort_values(self, column: str | list[str] | dict[str, Literal['ASC', 'DESC']],
                    ascending: bool = True) -> TableView:
        """
        Return a new table with the sorted data

        You can pass:
        1) A column-name with the optional parameter ascending
        2) A list of column names, which will all be ordered ascending
        3) A dictionary with the column-name and the sorting order, for ex:
        {'col1': 'ASC', 'col2': 'DESC', 'col3': 'ASC'}

        :param column: str | list | dict
        :param ascending: bool, default True
        """
        if isinstance(column, str):
            order_by_query = f'{column} {"ASC" if ascending else "DESC"}'

        elif isinstance(column, list):
            order_by_query = f'{", ".join(column)}'

        elif isinstance(column, dict):
            cols = [f'{col} {sort_order}' for col, sort_order in column.items()]
            order_by_query = f'{", ".join(cols)}'

        else:
            raise TypeError(f'column parameter must be str, list, or dict, not: {type(column)}')

        view_name = f'_table_sorted_{self.name}_{get_random_name(size=10)}_'
        query = f"""
        SELECT ROW_NUMBER() OVER (ORDER BY {order_by_query}) AS _rowid_, {self._cols_as_str}
        FROM {self.name} 
        """
        return self._create_and_get_temp_view(view_name=view_name, query=query)

    def limit(self, n: int) -> TableView:
        """
        Return a new Table with n amount of rows

        :param n: int
        :return: TableView
        """
        view_name = f'_table_limit_{self.name}_{get_random_name(size=10)}_'
        query = f'SELECT {ROWID}, {self._cols_as_str} FROM {self.name} WHERE _rowid_ <= {n}'
        return self._create_and_get_temp_view(view_name=view_name, query=query)

    def _create_and_get_temp_view(self, view_name: str, query: str) -> TableView:
        """
        Create a temporary-view (gets auto deleted at the end of the session) and return a new TableView instance

        :param view_name: str
        :param query: str
        :return: TableView instance
        """
        if 'AS _rowid_' not in query:
            raise ValueError('Query must alias the rowid column as `_rowid_` for `iloc` to work.')

        create_temp_view(
            conn=self.conn,
            view_name=view_name,
            query=query,
            drop_if_exists=False
        )
        return TableView(
            conn=self.conn,
            cache=self._cache,
            name=view_name,
            created_query=query
        )

    def __iter__(self) -> Generator[TableRow, None, None]:
        """
        Yield rows from cursor
        """
        with self.conn as cursor:
            yield from cursor.execute(self.query)

    def _get_col(self, column: str) -> Column:
        """
        Get column object

        :param column:str, column name
        :return: Column
        :raise: InvalidColumnError
        """
        if column not in self.columns:
            raise InvalidColumnError(f'Column must be one of the following: {self._cols_as_str}')
        return getattr(self, column)

    def _set_column(self, column: str) -> None:
        """
        Create and cache column object

        :param column: str
        :return: None
        """
        col_obj = Column(conn=self.conn, cache=self._cache, table_name=self.name, col_name=column)
        self._column_items[column] = col_obj

        if not hasattr(self, column):  # to avoid overwriting existing attributes and methods
            setattr(self, column, col_obj)

    def _column_slice(self, columns: list[str]) -> TableView:
        """
        Return a TableView with the given columns

        :param columns: list of column names
        :return: TableView
        """
        view_name = f'_table_{self.name}_{get_random_name(size=10)}_'
        query = f'SELECT {ROWID}, {", ".join(columns)} FROM {self.name}'
        return self._create_and_get_temp_view(view_name=view_name, query=query)

    @overload
    def __getitem__(self, item: Expression) -> TableView:
        ...

    @overload
    def __getitem__(self, item: str) -> Column:
        ...

    @overload
    def __getitem__(self, item: list[str]) -> TableView:
        ...

    def __getitem__(self, item: Expression | str | list[str]) -> Column | TableView:
        """
        Get column object for given column name

        :param item: Expression, column-name, or list of columns
        :return: Column
        :raise: KeyError
        """
        if isinstance(item, Expression):
            return self.filter(item)

        if isinstance(item, str):
            try:
                return self._get_col(item)
            except InvalidColumnError:
                raise KeyError(f'No such Column: {item}, must be one of the following: {self._cols_as_str}')

        if isinstance(item, list):
            return self._column_slice(item)

        raise TypeError(f'Argument must be of type str or Expression. not: {type(item)}')

    def __getattribute__(self, item) -> Any:
        """ Get attribute """
        # for avoiding 'Unresolved attribute' warnings (in Pycharm), this somehow fixes it
        return super().__getattribute__(item)

    def __len__(self) -> int:
        """ Return amount of rows """
        return self.len

    def __hash__(self) -> int:
        """ Get hash value of Table """
        return hash(f'{self.name}')

    def _repr_df(self) -> DataFrame:
        """
        Get a sample of the table data
        
        This method is a helper for: __repr__(), and _repr_html_().
        It returns a sample of the table data, by default the first and last 5 rows,...

        note on top_rows and bottom_rows;
        by default pandas display's the first and last 5 rows,
        passing a list than len() > 10 will get the shrink effect, which tells the user
        that there is more data than meets the eye.

        :return: DataFrame
        """
        top_rows = 10
        bottom_rows = 10
        n = len(self)

        if n <= top_rows:  # shortcut for small dataframes
            return DataFrame(data=self, columns=self.columns)

        data = self.iloc[:top_rows] + self.iloc[-bottom_rows:]
        index = list(range(top_rows)) + list(range(n - bottom_rows, n))
        return DataFrame(index=index, data=data, columns=self.columns)

    def __repr__(self) -> str:
        """ Return table as a Pandas DataFrame """
        n_rows, n_cols = self.shape
        size_info = f'\n\n[{n_rows} rows x {n_cols} columns]'
        return self._repr_df().to_string(show_dimensions=False, max_rows=10) + size_info

    def _repr_html_(self) -> str:
        """ Return table in HTML """
        return self._repr_df().to_html(show_dimensions=False, max_rows=10)

    def equals(self, other) -> bool:
        """
        Check if the content of a given Table is the same
        """
        if not isinstance(other, Table):
            return False

        if self.shape != other.shape:
            return False

        if id(self) == id(other):  # shortcut for same object from another variable
            return True

        if any(a != b for a, b in zip(self, other, strict=True)):
            return False
        return True


class TableView(Table):
    """
    A class that represents a View in SQL (or a virtual table in Sqlite3)

    All the methods are the same as a regular table, except the columns' property which will
    filter the list of columns to not return the '_rowid_' column.

    The reason the filtering is necessary is simply because each table in Sqlite3 has a _rowid_ column
    that is hidden (meaning it doesn't appear in 'SELECT * FROM table', only if explicitly queried, for ex:
    'SELECT _rowid_, * FROM table'),
    But virtual tables (views, temporary tables, etc.) don't have a builtin '_rowid_' column,
    so what we do, is we create a column containing each row's index when creating the view,
    but when we do 'SELECT * FROM my_view' we want to filter out the '_rowid_' column,
    so instead we pass the columns' property which will automatically filter it out.
    """
    # noinspection PyMissingConstructor
    def __init__(self, conn: sqlite3.Connection, cache: Cache, name: str, created_query: str = None) -> None:
        """
        Initialize the Table object

        :param conn: sqlite3.Connection
        :param cache: Cache, instance of Cache
        :param name: str, table name
        :param created_query: str | None
        """
        self.conn = conn
        self._cache = cache
        self.name = name
        self._column_items: dict[str, Column] = {}
        self._created_query = created_query  # save the query used in creating the table-view for debugging

        for column in self.columns:
            self._set_column(column)

    @property
    def columns(self) -> list[str]:
        """
        Get list with column names

        note that since a table view
        """
        with self.conn as cursor:
            return [x[1] for x in cursor.execute(f"PRAGMA table_info('{self.name}')")
                    if not x[1].startswith('_rowid_')]



from __future__ import annotations

import pandas as pd
from pympler import asizeof

import sqlite3
import itertools
import random
import string
from pathlib import Path
from typing import Generator, Iterable, Any, TypeVar, TYPE_CHECKING

from .exceptions import ViewAlreadyExists

if TYPE_CHECKING:
    from .connection import Database
    from .column import Column

__all__ = [
    'col_iterator',
    'sort_iterable_with_none_values',
    'convert_type_to_sql',
    'sql_tuple',
    'get_random_name',
    'create_temp_view',
    'concat',
    'get_mb_size',
    'get_gb_size',
    'rename_duplicate_cols',
    'convert_db_to_sql',
    'convert_csvs_to_db',
    'convert_sql_to_db',
    'load_sql_to_sqlite'
]

PrimitiveTypes = str | int | float | bool | None
T = TypeVar("T")
TypeAny = TypeVar('TypeAny', bound=Any)


def col_iterator(db: Database, *, numeric_only: bool = False) -> Generator[Column, None, None]:
    """ Generator that yields all the columns (objects) from all tables """
    for _, table in db.items():
        for _, col in table.items():
            if numeric_only:
                if col.data_is_numeric():
                    yield col
            else:
                yield col


def sort_iterable_with_none_values(it: Iterable) -> list:
    """
    Sort an Iterable that contains None values

    The builtin sorted() function isn't able to sort an iterable that contains any null values
    """
    return sorted(it, key=lambda x: (x is not None, x))


def convert_type_to_sql(x: str | int | float | bool) -> str:
    """
    Function that takes a value (primitive type) and returns an SQL-compatible string

    if type is str -> "'x'"
    if type is int or float -> '3' | '3.2'
    if type is bool -> 'true' | 'false'

    :param x: str | int | float | bool
    :return: str
    """
    if isinstance(x, str):
        return repr(x)
    if isinstance(x, bool):  # above int because bool inherits from int
        return str(x).lower()
    if isinstance(x, (int, float)):
        return str(x)

    raise TypeError(f'param x must be of type str, int, float, or bool. not: {type(x)}')


def sql_tuple(it: Iterable) -> str:
    """
    Convert an iterable to an SQL-compatible tuple

    :param it: Iterable
    :return: str
    """
    return f'({", ".join(convert_type_to_sql(x) for x in it)})'


def get_random_name(size: int = 10) -> str:
    """
    Get a string with random letters (from string.ascii_lowercase)

    :param size: int, default 10
    :return: str
    """
    return ''.join(random.choices(string.ascii_lowercase, k=size))


def create_temp_view(conn: sqlite3.Connection, view_name: str, query: str, drop_if_exists: bool = False) -> None:
    """
    Create temporary view from given sql query

    :param conn: sqlite3 connection
    :param view_name: str
    :param query: str, select query
    :param drop_if_exists: bool, default False
    :raises: ValueError if view_name already exists
    :return: None
    """
    with conn as cursor:
        views: Iterable[str] = (x[0] for x in cursor.execute("SELECT name FROM sqlite_temp_master WHERE type='view'"))

    if view_name in views:
        if drop_if_exists:
            with conn as cursor:
                cursor.execute(f'DROP VIEW {view_name}')
        else:
            raise ViewAlreadyExists(f"view {view_name} already exists")

    with conn as cursor:
        cursor.execute(f"CREATE TEMP VIEW {view_name} AS {query}")


def concat(*args: str | Iterable, sep: str = '') -> Generator[str, None, None]:
    """
    Return a generator with the elements concatenated (as a string)

    You can pass both strings and Iterables (list, tuple, set, dict, generator, etc..)
    if you pass an iterable than the length must be the same as the length of the column

    Example:
    it = concat(db.table.first_name, '-', db.table.last_name)
    print(next(it))
    # out: 'Jake-Roberts'

    :param args: str | Iterable
    :param sep: str, default: ''
    :return: Generator
    """
    converted_args = []
    for arg in args:  # make all args iterable, so we can do: `zip(*args)`
        if isinstance(arg, str) or not isinstance(arg, Iterable):
            arg = itertools.repeat(arg)
        converted_args.append(arg)

    for tup in zip(*converted_args):
        stringify_tup = map(str, tup)
        concat_tup = sep.join(stringify_tup)
        yield concat_tup


def get_mb_size(*obj) -> float:
    """
    A helper for getting the number of Megabytes an object/s is taking in memory

    :param obj: args, any object/s
    :return: float
    """
    bytes_size = asizeof.asizeof(*obj)
    return bytes_size / 1e+6


def get_gb_size(*obj) -> float:
    """
    A helper for getting the number of Gigabytes an object/s is taking in memory

    :param obj: args, any object/s
    :return: float
    """
    bytes_size = asizeof.asizeof(*obj)
    return bytes_size / 1e+9


def rename_duplicate_cols(columns: list) -> list:
    """
    for each duplicated column it will add a number as the suffix

    ex: ['a', 'b', 'c', 'a', 'b', 'b'] -> ['a', 'b', 'c', 'a_2', 'b_2', 'b_3']

    :param columns: DataFrame
    :return: list
    """
    new_cols = []
    prev_cols = []  # previously iterated columns in for loop

    for col in columns:
        prev_cols.append(col)
        count = prev_cols.count(col)

        if count > 1:
            new_cols.append(f'{col}_{count}')
        else:
            new_cols.append(col)
    return new_cols


def convert_db_to_sql(db_file: str | Path, sql_file: str | Path) -> None:
    """
    takes a .db file and converts it to .sql

    :param db_file: str, path/name to save new .db file
    :param sql_file: str, path to .sql file
    :return:
    """
    with sqlite3.connect(db_file) as conn:
        with open(sql_file, 'w') as file:
            for line in conn.iterdump():
                file.write(line)


def convert_csvs_to_db(db_file: str | Path, csv_files: list[str] | list[Path], set_lowercase: bool = True,
                       **kwargs: Any) -> None:
    """
    convert a list of CSVs to a database (.db file)

    if the param set_lowercase is true, it will rename the table and column names to lowercase
    note that any column name that contains spaces or dashes will replace the characters with underscores,
    for ex: 'first name' -> 'first_name'

    :param db_file: str, path/name to save new .db file
    :param csv_files: list, ex: ['orders.csv', 'names.csv', 'regions.csv'...]
    :param set_lowercase: bool, default True
    :param kwargs: key-word arguments to pass to pd.read_csv()
    :return: None
    """
    with sqlite3.connect(db_file) as conn:

        for csv in csv_files:
            df = pd.read_csv(csv, **kwargs)
            df.columns = df.columns.str.replace(' ', '_').str.replace('-', '_')

            if set_lowercase:
                df.columns = df.columns.str.lower()

            name = Path(csv).stem.replace(' ', '_').replace('-', '_')
            df.to_sql(name=name, con=conn, index=False)


def convert_sql_to_db(sql_file: str | Path, db_file: str | Path) -> None:
    """
    Convert .sql to .db file

    :param sql_file: str, path to .sql file
    :param db_file: str, path/name to save new .db file
    :return: None
    """
    with open(sql_file, 'r') as file:
        with sqlite3.connect(db_file) as conn:
            conn.executescript(file.read())


def load_sql_to_sqlite(sql_file: str | Path) -> sqlite3.Connection:
    """
    Create a Sqlite3 connection with a .sql file

    :param sql_file: str, path to .sql file
    :return: sqlite3 connection
    """
    with open(sql_file, 'r') as file:
        conn = sqlite3.connect(':memory:', check_same_thread=False)
        conn.executescript(file.read())
        return conn



from __future__ import annotations

from pandas import DataFrame

import sqlite3
import warnings
from typing import Generator, Any
from threading import Thread
from pathlib import Path

from .utils import convert_sql_to_db, rename_duplicate_cols
from .table import Table
from .exceptions import FileTypeError, InvalidTableError, ConnectionClosedWarning
from .cache import Cache


class Database:
    """
    A class that represents a database, all the tables will be stored as attributes in the Database object,
    and the columns will be stored as attributes in they're respective tables
    You can have a look at the README here: https://github.com/shner-elmo/pandas-db/blob/master/README.md
    """
    def __init__(self, db_path: str, cache: bool = True, populate_cache: bool = False,
                 max_item_size: int = 2, max_dict_size: int = 100) -> None:
        """
        Initialize the Database object

        The only required parameter is db_path, which takes a database file of type: db, sql, or sqlite.

        There are optional parameters for managing the cache, whenever the user calls a method like: Column.min()
        the method will check if we already stored the output of the SQL query ("SELECT MIN(Column) FROM Table")
        for the method in the cache, if we did, then it will return the result directly from the cache,
        otherwise: run the SQL query and save the output in cache for next time.

        The optional parameters for managing the cache are:

        cache: if True it will store the results of each SQL query in a dictionary for next time

        populate_cache: if True it will call the methods for each column in each table in the database,
        when its initialized, this way when the user calls the method it will already be present in cache
        and ready to use.
        note that this can take some time for very large databases.

        max_item_size: the max size in Megabytes an item can take in cache (sql_query + query_output)

        max_dict_size: the max size of the whole cache-dictionary, if the dictionary reaches its max size,
        no item will be added.

        :param db_path: str, path to database
        :param cache: bool, default True
        :param populate_cache: bool, default True
        :param max_item_size: int, size in MB
        :param max_dict_size: int, size in MB
        """
        self.db_path = db_path  # save for repr()
        db_path = Path(db_path)
        self.name = db_path.name
        extension = db_path.suffix.lower()
        valid_extension = ('.sql', '.db', '.sqlite', '.sqlite3')

        if extension not in valid_extension:
            raise FileTypeError(f'File extension must be one of the following: {", ".join(valid_extension)}')

        if extension == '.sql':  # if .sql load to .db file and then connect
            local_db_folder = Path(__file__).parent.parent / 'pandasdb-local-databases'
            local_db_file = local_db_folder / f'{db_path.stem}.db'

            if local_db_folder.exists():
                for f in local_db_folder.iterdir():
                    if f == local_db_file:
                        f.unlink()  # delete file and create a new one to avoid using old data
            else:
                local_db_folder.mkdir()

            convert_sql_to_db(sql_file=db_path, db_file=local_db_file)
            self.conn = sqlite3.connect(local_db_file, check_same_thread=False)
        else:
            self.conn = sqlite3.connect(db_path, check_same_thread=False)

        self.cache = Cache(
            conn=self.conn,
            cache_output=cache,
            max_item_size=max_item_size,
            max_dict_size=max_dict_size
        )

        self._table_items: dict[str, Table] = {}
        for table in self.tables:
            self._set_table(table=table)

        if cache and populate_cache:
            threads = []
            for _, table in self.items():
                thread = Thread(target=self.cache.populate_table, kwargs={'table': table})
                thread.start()
                threads.append(thread)

            for thread in threads:
                thread.join()

    @property
    def conn_open(self) -> bool:
        """
        Check if the SQL connection is open
        """
        try:
            self.conn.cursor()
            return True
        except sqlite3.ProgrammingError:
            return False

    @property
    def tables(self) -> list[str]:
        """
        Get list of tables

        :return: list with table names
        """
        with self.conn as cursor:
            return [x[0] for x in cursor.execute("SELECT name FROM sqlite_master WHERE type='table'")]

    @property
    def views(self) -> list[str]:
        """
        Get a list of all views

        Note that the filter method creates custom views for its filters but those all start with a double underscores.

        :return: list with table names
        """
        with self.conn as cursor:
            return [x[0] for x in cursor.execute("SELECT name FROM sqlite_master WHERE type='view'")]

    @property
    def temp_tables(self) -> list[str]:
        """
        Get a list of all temporary tables (a temporary table is a table that lasts only as long as the session does)
        """
        with self.conn as cursor:
            return [x[0] for x in cursor.execute("SELECT name FROM sqlite_temp_master WHERE type='table'")]

    @property
    def temp_views(self) -> list[str]:
        """
        Get a list of all temporary views (a temporary view is a table that lasts only as long as the session does)
        """
        with self.conn as cursor:
            return [x[0] for x in cursor.execute("SELECT name FROM sqlite_temp_master WHERE type='view'")]

    def get_columns(self, table_name: str) -> list[str]:
        """
        Get list of all columns within given table

        :param table_name: str
        :return: list with column names
        """
        if table_name not in self.tables + self.views:
            raise InvalidTableError(f'No such table: {table_name}')

        return [x[1] for x in self.cache.execute(f"PRAGMA table_info('{table_name}')")]

    def items(self) -> Generator[tuple[str, Table], None, None]:
        """
        Generator that yields: (table_name, table_object)
        """
        yield from self._table_items.items()

    def query(self, sql_query: str, rename_duplicates: bool = True) -> DataFrame:
        """
        Return a DataFrame with the query results

        Returns a DataFrame with the results of a given query,
        if there are columns with duplicated names it will add a number at the end, for ex:
        ['a', 'b', 'c', 'a', 'b', 'b'] -> ['a', 'b', 'c', 'a_2', 'b_2', 'b_3']

        :param sql_query: str, SQL query
        :param rename_duplicates: bool, default: True
        :return: DataFrame
        """
        with self.conn as cursor:
            data = cursor.execute(sql_query)

        cols = [x[0] for x in data.description]

        if rename_duplicates:
            if len(set(cols)) != len(cols):
                cols = rename_duplicate_cols(cols)

        return DataFrame(data=data, columns=cols)

    def __enter__(self) -> 'Database':
        """
        Return the instance of Database

        :return: self
        """
        return self

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        """
        Safely close the Database connection

        :param exc_type:
        :param exc_val:
        :param exc_tb:
        :return: None
        """
        self.exit()

    def __del__(self) -> None:
        """
        Perform cleanup as soon as the object is deleted

        :return: None
        """
        if hasattr(self, 'conn') and self.conn_open:
            self.exit()

    def exit(self) -> None:
        """
        Close Database connection, this should be called when you're done using the connection

        :return: None
        """
        if self.conn_open:
            self.conn.close()
        else:
            warnings.warn('Connection already closed!', ConnectionClosedWarning)

    def _set_table(self, table: str) -> None:
        """
        Create and set table-object to instance

        Create table object and save it to the self._table_items dictionary,
        and if the table-name isn't an existing attribute or method; then set it as an attribute as well.
        The table objects can be accessed using __getitem__ (db[table_name]) or as an attribute (db.table_name).

        :param table: str
        :return: None
        """
        table_obj = Table(conn=self.conn, cache=self.cache, name=table)
        self._table_items[table] = table_obj

        if not hasattr(self, table):  # to avoid overwriting existing attributes and methods
            setattr(self, table, table_obj)

    def __getitem__(self, table: str) -> Table:
        """
        Get Table object for given table_name

        :param table: str, table name
        :raise: KeyError if key not found
        :return: Table
        """
        if table not in self.tables:
            raise KeyError(f'No such Table: {table}, must be one of the following: {", ".join(self.tables)}')

        if table not in self._table_items:  # if table was created after the instance:
            self._set_table(table=table)

        return self._table_items[table]

    def __getattribute__(self, item) -> Any:
        """ Get attribute """
        # for avoiding 'Unresolved attribute' warnings (in Pycharm), this somehow fixes it
        return super().__getattribute__(item)

    def __len__(self) -> int:
        """ Get the number of tables in the database """
        return len(self.tables)

    def __repr__(self) -> str:
        """ Get the string representation of the class instance """
        return __class__.__name__ + "(db_path='{}')".format(self.db_path)



from __future__ import annotations

from pandas import DataFrame, Series

import sqlite3
import itertools
from typing import Generator, Callable, Any, Sequence, TypeVar, Iterable, overload

from .expression import Expression
from .cache import Cache
from .utils import create_temp_view, get_random_name, sql_tuple, convert_type_to_sql

ColumnValue = str | int | float | bool | None
Numeric = int | float
T = TypeVar('T')
ROWID = '_rowid_ AS _rowid_'  # select rowid/oid/_rowid_ as _rowid_


class IndexLoc:
    def __init__(self, column: Column) -> None:
        self.col = column
        self.len = len(column)  # to avoid recomputing

    def index_abs(self, idx: int) -> int:
        """
        Return the absolute of an index

        if the given index is negative it will convert it to positive, for ex:
        if the given index is -1 and the length of the table is 371 the method will return 371

        :param idx: int
        :return: int
        """
        if idx < 0:
            return self.len + idx
        return idx

    def validate_index(self, idx: int) -> None:
        """
        Assert given index is above zero, and below or equal to length,
        else: raise IndexError

        :param idx: int, must be positive
        :raise IndexError: if not 0 <= index < length
        :return: None
        """
        if not 0 <= idx < self.len:
            raise IndexError(f'Given index out of range ({idx})')

    @overload
    def __getitem__(self, index: int) -> ColumnValue:
        ...

    @overload
    def __getitem__(self, index: slice | list) -> list[ColumnValue]:
        ...

    def __getitem__(self, index: int | slice | list) -> ColumnValue | list[ColumnValue]:
        """
        Get row/value at given index

        if an int is passed; then values at the given index will be returned
        if a slice or a list is passed; then it will return a list of elements/cells from the column

        Note that the index is always increased by one before being passed to the SQL query,
        as the rowid columns starts the count at 1 and not 0.

        :param index: int | slice | list
        :return: tuple | list | BaseTypes
        """
        if isinstance(index, int):
            index = self.index_abs(index)
            self.validate_index(index)
            index += 1

            with self.col.conn as cursor:
                query = f'SELECT {self.col.name} FROM {self.col.table} WHERE _rowid_ == {index}'
                return cursor.execute(query).fetchall()[0][0]

        if isinstance(index, slice):
            indices = index.indices(self.len)
            indexes = [idx + 1 for idx in range(*indices)]

            with self.col.conn as cursor:
                query = f'SELECT {self.col.name} FROM {self.col.table} WHERE _rowid_ IN {sql_tuple(indexes)}'
                return [tup[0] for tup in cursor.execute(query)]

        if isinstance(index, list):
            indexes = [self.index_abs(idx) for idx in index]
            for idx in indexes:
                self.validate_index(idx)
            indexes = [idx + 1 for idx in indexes]
            unique_indexes = set(indexes)

            with self.col.conn as cursor:
                rows = cursor.execute(
                    f'SELECT _rowid_, {self.col.name} FROM {self.col.table} '
                    f'WHERE _rowid_ IN {sql_tuple(unique_indexes)}')

            idx_row_mapping: dict[int, ColumnValue] = dict(rows)
            return [idx_row_mapping[idx] for idx in indexes]

        raise TypeError(f'Index must be of type: int, list, or slice. not: {type(index)}')


class Column:
    """
    An object that represents a column of a table within a DataBase
    """
    def __init__(self, conn: sqlite3.Connection, cache: Cache, table_name: str, col_name: str) -> None:
        """
        Initialize the Column object

        :param conn: sqlite3.Connection
        :param cache: Cache, instance of Cache
        :param table_name: str
        :param col_name: str
        """
        self.conn = conn
        self._cache = cache
        self.table = table_name
        self.name = col_name
        self.query = f'SELECT {col_name} FROM {table_name}'

    @property
    def type(self) -> type:
        """
        Get column Python data type, i.e: str, int or float

        :return: type, str | int | float
        """
        out = self._cache.execute(f'{self.query} WHERE {self.name} IS NOT NULL LIMIT 1')[0][0]
        return type(out)

    @property
    def sql_type(self) -> str:
        """
        Get the column SQL data type as a string

        Return a string with the SQL data type, some of the most common are:
        TEXT, INTEGER, REAL, FLOAT, TIMESTAMP, BPCHAR, VARCHAR(250), NUMERIC(10,2), etc.

        :return str, e.g., TEXT, INTEGER, REAL...
        """
        for row in self._cache.execute(f"PRAGMA table_info('{self.table}')"):
            if row[1] == self.name:
                return row[2]

    def data_is_numeric(self) -> bool:
        """
        Return True if the column data is of type int or float, else: False
        """
        return self.type in (int, float)

    @property
    def len(self) -> int:
        """
        Get the amount of rows/ cells in the column (including None values)
        """
        return self._cache.execute(f'SELECT COUNT(*) FROM {self.table}')[0][0]

    def count(self) -> int:
        """
        Get the amount of rows/ cells in the column (excluding None values)
        """
        return self._cache.execute(f'SELECT COUNT({self.name}) FROM {self.table}')[0][0]

    def null_count(self) -> int:
        """
        Get the amount of None values in column
        """
        return self._cache.execute(f'SELECT COUNT(*) FROM {self.table} WHERE {self.name} IS NULL')[0][0]

    def min(self) -> str | int | float:
        """
        Get the min value of the column
        """
        return self._cache.execute(f'SELECT MIN({self.name}) FROM {self.table}')[0][0]

    def max(self) -> str | int | float:
        """
        Get the max value of the column
        """
        return self._cache.execute(f'SELECT MAX({self.name}) FROM {self.table}')[0][0]

    def sum(self) -> Numeric:
        """
        Get the sum of all values within the column

        :raise TypeError: if column isn't of type int or float
        :return float
        """
        if not self.data_is_numeric():
            raise TypeError(f'Cannot get sum for Column of type {self.type}')

        return self._cache.execute(f'SELECT SUM({self.name}) FROM {self.table}')[0][0]

    def avg(self) -> Numeric:
        """
        Get the avg value of the column

        :raise TypeError: if column isn't of type int or float
        :return float
        """
        if not self.data_is_numeric():
            raise TypeError(f'Cannot get avg for Column of type {self.type}')

        return self._cache.execute(f'SELECT AVG({self.name}) FROM {self.table}')[0][0]

    def median(self) -> Numeric:
        """
        Get the median value of the column

        :raise TypeError: if column isn't of type int or float
        :return float
        """
        if not self.data_is_numeric():
            raise TypeError(f'Cannot get median for Column of type {self.type}')

        if self.null_count() == 0:
            col = self.sort_values()
        else:
            col = self.not_null().sort_values()

        n = len(col)
        mid = (n - 1) // 2
        if n % 2 == 0:
            lst: list[Numeric] = col.iloc[[mid, mid + 1]]
            return sum(lst) / 2
        else:
            return col.iloc[mid]

    def mode(self) -> dict[ColumnValue, int]:
        """
        Get the mode/s of the column as a dictionary; {'value': count}

        :return dict
        """
        query = f"""
        SELECT {self.name}, COUNT(*) FROM {self.table}
        GROUP BY 1
        HAVING COUNT(*) >= (	
            SELECT COUNT(*) FROM {self.table}
            GROUP BY {self.name}
            ORDER BY 1 DESC
            LIMIT 1
        )
        """
        return dict(self._cache.execute(query))

    def describe(self) -> dict[str, str | int | float]:
        """
        Get a dictionary with different properties for the column

        if column data is numeric return a dictionary with keys:
        {'len', 'count', 'min', 'max', 'sum', 'avg', 'median'}
        if its text data:
        {'len', 'count', 'min', 'max', 'mode'}

        :return dict
        """
        if self.data_is_numeric():
            return {
                'len': self.len,
                'count': self.count(),
                'min': self.min(),
                'max': self.max(),
                'sum': self.sum(),
                'avg': self.avg(),
                'median': self.median()
            }
        else:
            return {
                'len': self.len,
                'count': self.count(),
                'min': self.min(),
                'max': self.max(),
                'unique': len(self.unique())
            }

    def unique(self) -> list[ColumnValue]:
        """
        Get list with unique values

        :return list
        """
        return [tup[0] for tup in self._cache.execute(f'SELECT DISTINCT {self.name} FROM {self.table}')]

    def value_counts(self) -> dict[ColumnValue, int]:
        """
        Get a dictionary with the count of each value in the Column

        example:
        column = ['a', 'b', 'c', 'b', 'c', 'b'] -> {'a': 1, 'b': 3, 'c': 2}

        :return: dict
        """
        query = f"""
        SELECT {self.name}, COUNT(*) FROM {self.table}
        WHERE {self.name} IS NOT NULL
        GROUP BY 1
        ORDER BY 2 DESC, 1 ASC
        """
        return dict(self._cache.execute(query))

    def to_series(self) -> Series:
        """
        Return column as a Pandas Series

        :return Pandas Series
        """
        return Series(data=iter(self), name=self.name)

    def data(self, limit: int = None) -> list[ColumnValue]:
        """
        Get column-data

        If limit is None: return all data, else: return n_amount of rows/ cells

        :param limit: int
        :return: list
        """
        with self.conn as cursor:
            if limit:
                return [tup[0] for tup in cursor.execute(self.query + f' LIMIT {limit}')]
            return [tup[0] for tup in cursor.execute(self.query)]

    def sample(self, n: int = 10) -> list[ColumnValue]:
        """
        Get a list of random values from the column

        :param n: int, number of values
        :return: list
        """
        with self.conn as cursor:
            return [tup[0] for tup in cursor.execute(f'{self.query} ORDER BY RANDOM() LIMIT {n}')]

    def apply(self, func: Callable[[ColumnValue, ...], T], *, ignore_na: bool = True,
              args: tuple = (), **kwargs: Any) -> Generator[T, None, None]:
        """
        Apply function on each cell in the column

        example:
        db = DataBase('data/parch-and-posey.sql')
        column = db.accounts.primary_poc.apply(lambda x: x.split()[0])
        for first_name in column:
            print(first_name)

        'Tamara'
        'Sung'
        'Jodee'
        'Serafina'

        :param func: Callable
        :param ignore_na: bool, default: True
        :param args: tuple, args to pass to the function
        :param kwargs: keyword arguments to pass to the function
        :return: Generator
        """
        for cell in self:
            if cell is None and ignore_na:
                yield cell
            else:
                yield func(cell, *args, **kwargs)

    @property
    def iloc(self) -> IndexLoc:
        """
        Get data by: index, list, or slice

        Getitem supports three ways of indexing the iterable:
        1) Singular Integer, ex: IndexIloc[0], IndexIloc[32], or with negative: IndexIloc[-12]
        2) Passing a list of integers, ex: IndexIloc[[1, 22, 4, 3, 17, 38]], IndexIloc[[1, -4, 17, 22, 38, -4, -1]]
        4) Passing Slice, ex: IndexIloc[:10], IndexIloc[2:8], IndexIloc[2:24:2]

        The return type will be a list for multiple items,
        and one of the following: str, int, or float. Depending on the data type of the column

        :return: list, str, int, or float
        """
        return IndexLoc(self)

    def not_null(self) -> ColumnView:
        """
        Return a new column without any null values

        :return: ColumnView
        """
        return self.filter(expression=Expression(query=f'{self.name} IS NOT NULL', table=self.table))

    def sort_values(self, ascending: bool = True) -> ColumnView:
        """
        Return a new column with the data sorted

        :param ascending: bool, default True
        :return: ColumnView
        """
        view_name = f'_col_sorted_{self.table}_{self.name}_{get_random_name(size=10)}_'
        query = f"""
        SELECT
            ROW_NUMBER() OVER (ORDER BY {self.name} {'ASC' if ascending else 'DESC'}) AS _rowid_, 
            {self.name}
        FROM {self.table} 
        """
        return self._create_and_get_temp_view(view_name=view_name, query=query)

    def limit(self, n: int) -> ColumnView:
        """
        Return a new column limits the amount of rows in a column

        (creates a view with: f"SELECT ... LIMIT {n}")

        :param n: int
        :return: ColumnView
        """
        view_name = f'_col_sorted_{self.table}_{self.name}_{get_random_name(size=10)}_'
        query = f"SELECT {ROWID}, {self.name} FROM {self.table} WHERE _rowid_ <= {n}"
        return self._create_and_get_temp_view(view_name=view_name, query=query)

    def filter(self, expression: Expression) -> ColumnView:
        """
        Return a new Column object with the filtered data

        - How it works:

        It takes an Expression, which could be either as an instance of the Expression class, for ex:
        `db.table.country.filter(Expression(query='country IN ("USA", "JPY", "ITA")'))`

        or from using an operator on a Column object which will return an instance of Expression, for ex:
        `col = db.table.col`
        `col.filter(col > 35.93)` # return a new column with only records which have `col` above 35.93

        or using Column.isin():
        `countries = db.countries`
        # return a column showing the population for the countries: USA, JPY, and ITA.
        `countries.population.filter(countries.isin(("USA", "JPY", "ITA")))`

        You can also pass the expression with square brackets to the column:
        `col = db.table.col`
        `col[col > 35.93]`

        Once the expression is passed, an SQL view will be created with the query being:
        "SELECT {col} FROM {table} WHERE {expression}"
        and then a new Column instance is returned with the name being the view-name

        --------------------------------------------------------------------------------------------------------
        the view name will start with '_col_` since the data within the table-view represents a column,
        and we add 10 random letters to the name, so we can create other filters for the same column without
        having to delete previous ones, or overwrite them.

        note that the created view is temporary, meaning that when the user closes the connection
        with `db.exit()`, all the views in the list will be dropped/ deleted automatically.

        A final note on 'ROW_NUMBER' in the SQL query;
        for the `Column.iloc` to work we need to have a rowid column, sqlite3 has it builtin on each table,
        meaning its auto-generated (but the user can overwrite it),
        the issue is that since a VIEW is a virtual-table it doesn't have the rowid column.
        so we are left with one option; select the `rowid` column from the table when creating the view,
        but the issue with that is that the position of the rows no longer corresponds since the amount of rows
        in each table is different,
        so what I did instead is, create a column using the `ROW_NUMBER()` function to get the index/id of each row,
        so it will order the table by rowid, and then create the column using the `ROW_NUMBER` function,
        and finally, alias the new column as '_rowid_' so `self.iloc` can reference it and use it to get rows
        by index position.

        :param expression: Expression
        :return: ColumnView instance
        """
        view_name = f'_col_filtered_{self.table}_{self.name}_{get_random_name(size=10)}_'
        query = f"""
        SELECT ROW_NUMBER() OVER (ORDER BY _rowid_) AS _rowid_, {self.name}
        FROM {self.table} 
        WHERE {expression.query}
        """
        return self._create_and_get_temp_view(view_name=view_name, query=query)

    # def astype(self, py_type) -> ColumnView:
    #     """
    #     Convert/ cast data to given type (if possible)
    #
    #     :param py_type: data-type: str, int, float
    #     :return: ColumnView
    #     """
    #     # sqlite3 data types: https://www.sqlite.org/datatype3.html
    #     valid_types = (str, int, float, bool)
    #     if py_type not in valid_types:
    #         raise ValueError(f'Type must be one of the following: {valid_types}')
    #
    #     py_to_sql_types = {str: 'TEXT', int: 'INTEGER', float: 'REAL', bool: 'BOOLEAN'}
    #     return self._create_and_get_temp_view(
    #         view_name=f'_col_casted_{self.table}_{self.name}_{get_random_name(size=10)}_',
    #         query=f'SELECT CAST({self.name} AS {py_to_sql_types[py_type]}, {ROWID} FROM {self.table}'
    #     )

    def _create_and_get_temp_view(self, view_name: str, query: str) -> ColumnView:
        """
        Create a temporary-view (gets auto deleted at the end of the session) and return a new ColumnView instance

        :param view_name: str
        :param query: str
        :return: ColumnView instance
        """
        if 'AS _rowid_' not in query:
            raise ValueError('Query must alias the rowid column as `_rowid_` for `iloc` to work.')

        create_temp_view(
            conn=self.conn,
            view_name=view_name,
            query=query,
            drop_if_exists=False
        )
        return ColumnView(
            conn=self.conn,
            cache=self._cache,
            table_name=view_name,
            col_name=self.name,
            created_query=query
        )

    def __getitem__(self, item: int | slice | list | Expression) -> Any:
        """
        Return index slice or filtered Column

        You can do two things with Column.__getitem__():
        1. Get value/s at given index
        2. Get a filtered Column

        There are three ways to get a value or list of values at a given index:
        1. pass an integer: db.table.column[28]
        2. passing a slice: db.table.column[8:24:2]
        3. and using a list of indexes: db.table.column[[3, 2, 8, -1, 15]]

        And for filtering a column you can simply pass a column with a logical expression:
        col1 = db.table.col1
        col1[col1 > 10]

        :param item: int | slice | list | Expression | tuple
        :return: IndexLoc | Column
        """
        if isinstance(item, Expression):
            return self.filter(item)
        if isinstance(item, (int, slice, list)):
            return self.iloc[item]

        raise TypeError(f'Argument must be of type Expression, int, slice, or list. not: {type(item)}')

    def __iter__(self) -> Generator[ColumnValue, None, None]:
        """ Yield values from column """
        with self.conn as cursor:
            for i in cursor.execute(self.query):
                yield i[0]

    def __len__(self) -> int:
        """ Get amount of rows """
        return self.len

    def __hash__(self) -> int:
        """ Get hash value of Column """
        return hash(f'{self.table}.{self.name}')

    def _repr_df(self) -> DataFrame:
        """
        Convert column to Dataframe

        Convert column to a Dataframe but only first and last five rows,
        without iterating through each row.

        :return: DataFrame
        """
        top_rows = 10
        bottom_rows = 10
        n = len(self)

        if n <= top_rows:  # shortcut for small dataframes
            return DataFrame(data=self, columns=[self.name])

        data = self.iloc[:top_rows] + self.iloc[-bottom_rows:]
        index = list(range(top_rows)) + list(range(n - bottom_rows, n))
        return DataFrame(index=index, data=data, columns=[self.name])

    def __repr__(self) -> str:
        """ Return column as a Pandas Series """
        return self._repr_df().to_string(show_dimensions=False, max_rows=10)

    def _repr_html_(self) -> str:
        """ Return column in HTML """
        return self._repr_df().to_html(show_dimensions=False, max_rows=10)

    def __contains__(self, item: ColumnValue) -> bool:
        """
        Return True if item is present in the column, else False
        """
        with self.conn as cur:
            out = cur.execute(f'{self.query} WHERE {self.name} = {convert_type_to_sql(item)} LIMIT 1')
        return len(out.fetchall()) == 1

    def __add__(self, other: Column | Iterable | Numeric | str) -> Generator[ColumnValue, None, None]:
        """
        Return a generator with the arithmetic operation applied on each element

        :param other: Column | Iterable | int | float | bool | str
        :return: Generator
        """
        if isinstance(other, str) or not isinstance(other, Iterable):
            other = itertools.repeat(other, len(self))

        for x, y in zip(self, other, strict=True):
            if x is None:
                yield x
            else:
                yield x + y

    def __sub__(self, other: Column | Iterable | Numeric) -> Generator[ColumnValue, None, None]:
        """
        Return a generator with the arithmetic operation applied on each element

        :param other: Column | Iterable | float | int
        :return: None
        """
        if not isinstance(other, Iterable):
            other = itertools.repeat(other, len(self))

        for x, y in zip(self, other, strict=True):
            if x is None:
                yield x
            else:
                yield x - y

    def __mul__(self, other: Column | Iterable | Numeric) -> Generator[ColumnValue, None, None]:
        """
        Return a generator with the arithmetic operation applied on each element

        :param other: Column | Iterable | float | int
        :return: None
        """
        if not isinstance(other, Iterable):
            other = itertools.repeat(other, len(self))

        for x, y in zip(self, other, strict=True):
            if x is None:
                yield x
            else:
                yield x * y

    def __truediv__(self, other: Column | Iterable | Numeric) -> Generator[ColumnValue, None, None]:
        """
        Return a generator with the arithmetic operation applied on each element

        :param other: Column | Iterable | float | int
        :return: None
        """
        if not isinstance(other, Iterable):
            other = itertools.repeat(other, len(self))

        for x, y in zip(self, other, strict=True):
            if x is None:
                yield x
            else:
                yield x / y

    def __floordiv__(self, other: Column | Iterable | Numeric) -> Generator[ColumnValue, None, None]:
        """
        Return a generator with the arithmetic operation applied on each element

        :param other: Column | Iterable | float | int
        :return: None
        """
        if not isinstance(other, Iterable):
            other = itertools.repeat(other, len(self))

        for x, y in zip(self, other, strict=True):
            if x is None:
                yield x
            else:
                yield x // y

    def __gt__(self, other: Numeric) -> Expression:
        """

        :param other: float
        :return: Expression
        """
        return Expression(query=f'{self.name} > {convert_type_to_sql(other)}', table=self.table)

    def __ge__(self, other: Numeric) -> Expression:
        """

        :param other: float
        :return: Expression
        """
        return Expression(query=f'{self.name} >= {convert_type_to_sql(other)}', table=self.table)

    def __lt__(self, other: Numeric) -> Expression:
        """

        :param other: float
        :return: Expression
        """
        return Expression(query=f'{self.name} < {convert_type_to_sql(other)}', table=self.table)

    def __le__(self, other: Numeric) -> Expression:
        """

        :param other: float
        :return: Expression
        """
        return Expression(query=f'{self.name} <= {convert_type_to_sql(other)}', table=self.table)

    def __eq__(self, other: ColumnValue) -> Expression:
        """

        :param other: str or float
        :return: Expression
        """
        if other is None:
            return Expression(query=f'{self.name} IS NULL', table=self.table)
        return Expression(query=f'{self.name} = {convert_type_to_sql(other)}', table=self.table)

    def __ne__(self, other: ColumnValue) -> Expression:
        """

        :param other: str or float
        :return: Expression
        """
        if other is None:
            return Expression(query=f'{self.name} IS NOT NULL', table=self.table)
        return Expression(query=f'{self.name} != {convert_type_to_sql(other)}', table=self.table)

    def isin(self, options: Sequence) -> Expression:
        """

        :param options: tuple
        :return: Expression
        """
        options = sql_tuple(options)
        return Expression(query=f'{self.name} IN {options}', table=self.table)

    def between(self, a: Numeric, b: Numeric) -> Expression:
        """

        :param a: float
        :param b: float
        :return: Expression
        """
        return Expression(query=f'{self.name} BETWEEN {convert_type_to_sql(a)} AND {convert_type_to_sql(b)}',
                          table=self.table)

    def like(self, regex: str) -> Expression:
        """

        :param regex: str
        :return: Expression
        """
        return Expression(query=f'{self.name} LIKE {convert_type_to_sql(regex)}', table=self.table)

    # SQLite3 doesn't support ILIKE (LIKE is already case-insensitive)
    # def ilike(self, regex: str) -> Expression:
    #     """
    #
    #     :param regex: str
    #     :return: Expression
    #     """
    #     return Expression(query=f"{self.name} ILIKE '{regex}'", table=self.table)


class ColumnView(Column):
    """
    A ColumnView is created everytime we filter an existing Column.
    (from Column.filter() or Column[<expression>]
    """
    # noinspection PyMissingConstructor
    def __init__(self, conn: sqlite3.Connection, cache: Cache,
                 table_name: str, col_name: str, created_query: str = None) -> None:
        """
        Initialize the Column object

        :param conn: sqlite3.Connection
        :param cache: Cache, instance of Cache
        :param table_name: str
        :param col_name: str
        :param created_query: None | str
        """
        self.conn = conn
        self._cache = cache
        self.table = table_name
        self.name = col_name
        self.query = f'SELECT {col_name} FROM {table_name}'
        self._created_query = created_query  # save the query used in creating the column-view for debugging




class FileTypeError(Exception):
    """ Raised when a given file type isn't valid """
    pass


class ViewAlreadyExists(Exception):
    """ Raised when the user tries to create an SQL View that already exists """
    pass


class InvalidTableError(Exception):
    """ Raised when requested table isn't present in the DataBase """
    pass


class InvalidColumnError(Exception):
    """ Raised when requested column isn't present in the DataBase """
    pass


class ExpressionError(Exception):
    """ Raised when trying to use an instance of Expression with a non-Expression instance """
    pass


class ConnectionClosedWarning(Warning):
    """ Raised when using a closed SQL connection """
    pass


class DatabaseError(Exception):
    """ Base Database Error """
    pass



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pandasdb2']

package_data = \
{'': ['*']}

install_requires = \
['pandas>=1.5.2,<2.0.0', 'pympler>=1.0.1,<2.0.0']

setup_kwargs = {
    'name': 'pandasdb2',
    'version': '2.0.2',
    'description': 'Pandas but without bringing the data into memory',
    'long_description': '# Pandas-DB\n\n### A lightweight object for analyzing data directly from a Database without having to load anything onto memory\n\n---\n[![PyPi](https://img.shields.io/badge/PyPi-2.0.1-yellow)](https://pypi.org/project/pandasdb2/)\n[![Downloads](https://pepy.tech/badge/pandasdb2)](https://pepy.tech/project/pandasdb2)\n[![Downloads](https://pepy.tech/badge/pandasdb2/month)](https://pepy.tech/project/pandasdb2)\n\nYou can get the package directly from [PyPI](https://pypi.org/project/pandasdb2/)\n```\npip install pandasdb2\n```\n---\n## Why use Pandas-DB?\n\nPandas-db is a read-only package that allows you to view, analyze and explore all the content in a given Database (supported file extensions: db, sql, and sqlite)\n\nThe advantage of using this over something like Pandas is that your not storing anything onto memory and therefore the `db` object is very light compared to a regular Pandas `DataFrame` \n\n### How does it work?\nThe instance only stores the names of the tables as attributes (which is great for auto-complete), the Table and Column objects as the attribute values, and of course a reference to the SQLite connection object. \n\nSo whenever you call a method to get some data, for ex: `db.table.data(10)` or `db.column.describe()`;  \nthe function runs a query that gets the data directly from the Database, so there is no need to store anything in memory.  \n\nYou can think of it as a wrapper for SQL so there is no need to type SQL queries manually, but also has some of the most common methods and functions from Pandas so you also don\'t need to store the table data onto memory. \n\nAnd to top it off, it makes it easy to import the tables from a database onto pandas for further analysis, for ex:\n```python\nfrom pandasdb import Database\ndb = Database(db_path=\'data/parch-and-posey.sql\')\ndb.tables\n```\n```\n[\'web_events\', \'sales_reps\', \'region\', \'orders\', \'accounts\']\n```\n\n```\ndf1 = db.orders.to_df()\ndf2 = db.accounts.to_df()\ntype(df1), type(df2)\n```\n```\n(pandas.core.frame.DataFrame, pandas.core.frame.DataFrame)\n```\n\n\n---\n## Memory Usage\n\nNow let\'s visualize this and see the amount of memory a `Database` object consumes compared to a Pandas `DataFrame`\n\nImport the package:\n```python\nfrom pandasdb import Database\n```\nCreate a connection:\n```python\ndb = Database(db_path=\'.../yfin_data.db\')\n```\n\nFor example, this Database contains one table with more than 18 million rows:\n```python\ndb.tables, db.stock_data.shape\n```\n```\n([\'stock_data\'], (18078890, 9))\n```\n\n\n`db` object size:\n```python\nfrom pandasdb.utils import get_gb_size\nget_gb_size(db)\n```\n```\n0.000005 GB\n```\n\n`dataframe` object size:\n```python\ndataframe = db.stock_data.to_df()\nget_gb_size(dataframe)\n```\n```\n6.961891 GB\n```\nAs you can see, while the Database object remains almost the same size the Dataframe increases exponentially.\n\n---\n\n### Now let\'s dive onto the pandasdb package\n\nOnce you have the package installed, import the package:\n```python\nfrom pandasdb import Database\n```\n\nTo instantiate the Database class you need to pass the path to the Database file, which could be one of the following extensions: db, sql, or sqlite\n```python\ndb = Database(db_path=\'data/forestation.db\')\n```\n\n### Tables\nGet a list of all the tables \n```python\ndb.tables\n```\n```\n[\'forest_area\', \'land_area\', \'regions\']\n```\n\nTo get the table object we can either use square brackets or type the name as an attribute, just like in Pandas\n```python\ndb[\'forest_area\']\n```\n```\n     country_code  country_name  year  forest_area_sqkm\n0             ABW         Aruba  2016           4.20000\n1             AFG   Afghanistan  2016       13500.00000\n2             AGO        Angola  2016      577311.99220\n3             ALB       Albania  2016        7705.39978\n4             AND       Andorra  2016         160.00000\n           ...           ...   ...               ...\n5881          XKX        Kosovo  1990               NaN\n5882          YEM   Yemen, Rep.  1990        5490.00000\n5883          ZAF  South Africa  1990       92410.00000\n5884          ZMB        Zambia  1990      528000.00000\n5885          ZWE      Zimbabwe  1990      221640.00000\n[5886 rows x 4 columns]\n```\n\nOr:\n```python\ndb.forest_area\n```\n```\n     country_code  country_name  year  forest_area_sqkm\n0             ABW         Aruba  2016           4.20000\n1             AFG   Afghanistan  2016       13500.00000\n2             AGO        Angola  2016      577311.99220\n3             ALB       Albania  2016        7705.39978\n4             AND       Andorra  2016         160.00000\n           ...           ...   ...               ...\n5881          XKX        Kosovo  1990               NaN\n5882          YEM   Yemen, Rep.  1990        5490.00000\n5883          ZAF  South Africa  1990       92410.00000\n5884          ZMB        Zambia  1990      528000.00000\n5885          ZWE      Zimbabwe  1990      221640.00000\n[5886 rows x 4 columns]\n```\n\nGet a list of the table columns\n```python\ndb.forest_area.columns\n```\n```\n[\'country_code\', \'country_name\', \'year\', \'forest_area_sqkm\']\n```\n\nGet table shape:\n```python\ndb.forest_area.shape\n```\n```\n(5886, 4)\n```\n\nGet the first 20 rows \n```python\ndb.forest_area.data(20)\n```\n```\n[(\'ABW\', \'Aruba\', 2016, 4.199999869),\n (\'AFG\', \'Afghanistan\', 2016, 13500.0),\n (\'AGO\', \'Angola\', 2016, 577311.9922),\n (\'ALB\', \'Albania\', 2016, 7705.39978),\n (\'AND\', \'Andorra\', 2016, 160.0),\n ...\n```\nOr you can use the `iloc` property to get a slice:\n```python\ndb.regions.iloc[5:10]\n```\n```\n[(\'Angola\', \'AGO\', \'Sub-Saharan Africa\', \'Lower middle income\'),\n (\'Antigua and Barbuda\', \'ATG\', \'Latin America & Caribbean\', \'High income\'),\n (\'Argentina\', \'ARG\', \'Latin America & Caribbean\', \'High income\'),\n (\'Armenia\', \'ARM\', \'Europe & Central Asia\', \'Upper middle income\'),\n (\'Aruba\', \'ABW\', \'Latin America & Caribbean\', \'High income\')]\n```\n---\n\n### Columns:\n```python\ndb.regions.region\n```\n```\nOut[18]: \n0                      South Asia\n1           Europe & Central Asia\n2      Middle East & North Africa\n3             East Asia & Pacific\n4           Europe & Central Asia\n                  ...            \n214    Middle East & North Africa\n215    Middle East & North Africa\n216            Sub-Saharan Africa\n217            Sub-Saharan Africa\n218                         World\nName: region, Length: 219, dtype: object\n```\n\nGet the Python and SQL data type of the column:\n```python\ndb.regions.region.type, db.regions.region.sql_type\n```\n```\n(str, \'TEXT\')\n```\n\nDescribe the column: \n```python\ndb.forest_area.country_name.describe()\n```\n```\n{\'len\': 5886,\n \'count\': 5886,\n \'min\': \'Afghanistan\',\n \'max\': \'Zimbabwe\',\n \'unique\': 218}\n```\nNote that the len returns the number of rows/ items in the column, while count excludes None values\n\nDescribe a numeric column:\n```python\ndb.forest_area.forest_area_sqkm.describe()\n```\n```\n{\'len\': 5886,\n \'count\': 5570,\n \'min\': 0.799999982,\n \'max\': 41282694.9,\n \'sum\': 2178158753.6738772,\n \'avg\': 391051.84087502287,\n \'median\': 20513.00049}\n```\n\nGet a dictionary with each distinct value and its count:\n```python\ndb.regions.region.value_counts()\n```\n```\n{\'Europe & Central Asia\': 58,\n \'Sub-Saharan Africa\': 48,\n \'Latin America & Caribbean\': 42,\n \'East Asia & Pacific\': 38,\n \'Middle East & North Africa\': 21,\n \'South Asia\': 8,\n \'North America\': 3,\n```\n\nGet the most common value:\n```python\ndb.regions.income_group.mode()\n```\n```\n{\'High income\': 81}\n```\n\nGet all unique/ distinct values in a column:\n```python\ndb.forest_area.country_name.unique()\n```\n```\n[\'Aruba\',\n \'Afghanistan\',\n \'Angola\',\n ...\n \'Zambia\',\n \'Zimbabwe\']\n```\n\nWhen you\'re done you should always call the `exit()` method to close the SQL connection safely\n```python\ndb.exit()\n```\n\nNow let\'s use another Database\n```python\ndb = Database(db_path=\'data/parch-and-posey.sql\')\n\nfor table in db.tables:\n    print(table, db.get_columns(table_name=table))\n```\n```\nweb_events [\'id\', \'account_id\', \'occurred_at\', \'channel\']\nsales_reps [\'id\', \'name\', \'region_id\']\nregion [\'id\', \'name\']\norders [\'id\', \'account_id\', \'occurred_at\', \'standard_qty\', \'gloss_qty\', \'poster_qty\', \'total\', \'standard_amt_usd\', \'gloss_amt_usd\', \'poster_amt_usd\', \'total_amt_usd\']\naccounts [\'id\', \'name\', \'website\', \'lat\', \'long\', \'primary_poc\', \'sales_rep_id\']\n```\n\nYou can apply a function to the whole column like so: \n```python\ncolumn = db.accounts.primary_poc.apply(lambda x: x.split(\' \')[-1])\n\nfor last_name in column:\n    print(last_name)\n```\n```\nTuma\nShields\nLupo\nBanda\nCrusoe\n...\n```\n\nSimilarly, the Table object has an `applymap` method, which maps/ applies the function on each cell in the table\n\nFirst, let\'s have a look at the table:\n```python\ndb.sales_reps\n```\n```\n        id                 name  region_id\n0   321500        Samuel Racine          1\n1   321510         Eugena Esser          1\n2   321520      Michel Averette          1\n3   321530        Renetta Carew          1\n4   321540          Cara Clarke          1\n..     ...                  ...        ...\n45  321950         Elwood Shutt          4\n46  321960  Maryanna Fiorentino          4\n47  321970  Georgianna Chisholm          4\n48  321980       Micha Woodford          4\n49  321990          Dawna Agnew          4\n\n[50 rows x 3 columns]\n```\n\nAnd now we\'re going to pass a lambda function that will take the cell value and return the number of characters by converting it to a string:\n```python\ntable = db.sales_reps.applymap(lambda x: len(str(x)))\n\nfor row in table:\n    print(row)\n```\n```\n(6, 13, 1)\n(6, 12, 1)\n(6, 15, 1)\n(6, 13, 1)\n(6, 11, 1)\n...\n```\n\nYou can also iterate directly on the table/ column object: \n```python\nfor row in db.sales_reps:\n    print(row)\n```\n```\n(321500, \'Samuel Racine\', 1)\n(321510, \'Eugena Esser\', 1)\n(321520, \'Michel Averette\', 1)\n(321530, \'Renetta Carew\', 1)\n(321540, \'Cara Clarke\', 1)\n...\n```\n\nConvert a table to a `DataFrame`:\n```python\ndf = db.orders.to_df()\ntype(df)\n```\n```\npandas.core.frame.DataFrame\n```\n\nSimilarly, you can convert a column to a Pandas `Series`\n```python\nser = db.orders.occurred_at.to_series()\ntype(ser)\n```\n```\npandas.core.series.Series\n```\n\n\n\n### Filtering\n\nJust like Pandas you can filter a given table by passing a column object with an operator, eg: `col >= 32`\n```python\ndb = Database(\'data/parch-and-posey.sql\')\ndb.tables\n```\n```\n[\'web_events\', \'sales_reps\', \'region\', \'orders\', \'accounts\']\n```\n\nLets save the table that we\'ll be working with as df\n```python\ndf = db.orders\n```\n```\n        id  account_id          occurred_at  standard_qty  gloss_qty  poster_qty  total  standard_amt_usd  gloss_amt_usd  poster_amt_usd  total_amt_usd\n0        1        1001  2015-10-06 17:31:14           123         22          24    169            613.77         164.78          194.88         973.43\n1        2        1001  2015-11-05 03:34:33           190         41          57    288            948.10         307.09          462.84        1718.03\n2        3        1001  2015-12-04 04:21:55            85         47           0    132            424.15         352.03            0.00         776.18\n3        4        1001  2016-01-02 01:18:24           144         32           0    176            718.56         239.68            0.00         958.24\n4        5        1001  2016-02-01 19:27:27           108         29          28    165            538.92         217.21          227.36         983.49\n    ...         ...                  ...           ...        ...         ...    ...               ...            ...             ...            ...\n6907  6908        4501  2016-06-29 04:03:39            11        199          59    269             54.89        1490.51          479.08        2024.48\n6908  6909        4501  2016-07-29 19:58:32             5         91          96    192             24.95         681.59          779.52        1486.06\n6909  6910        4501  2016-08-27 00:58:11            16         94          82    192             79.84         704.06          665.84        1449.74\n6910  6911        4501  2016-11-22 06:52:22            63         67          81    211            314.37         501.83          657.72        1473.92\n6911  6912        4501  2016-12-21 13:30:42            61        150          52    263            304.39        1123.50          422.24        1850.13\n\n[6912 rows x 11 columns]\n```\n\nGet all the orders where the account-id is equal to 4091\n```python\ndf[df.account_id == 4091]\n```\n```\n     id  account_id          occurred_at  standard_qty  gloss_qty  poster_qty  total  standard_amt_usd  gloss_amt_usd  poster_amt_usd  total_amt_usd\n0  3775        4091  2016-11-22 07:33:05           325         36          46    407           1621.75         269.64          373.52        2264.91\n1  3776        4091  2016-12-21 09:22:20           300         28           0    328           1497.00         209.72            0.00        1706.72\n2  6573        4091  2016-11-22 07:57:27            45        482         305    832            224.55        3610.18         2476.60        6311.33\n\n[3 rows x 11 columns]\n```\n\nGet all the orders where the total quantity is between 300 - 500\n```python\ndf[df.total.between(300, 500)]\n```\n```\n        id  account_id          occurred_at  standard_qty  gloss_qty  poster_qty  total  standard_amt_usd  gloss_amt_usd  poster_amt_usd  total_amt_usd\n0       25        1041  2016-10-14 23:54:21           298         28          69    395           1487.02         209.72          560.28        2257.02\n1       26        1041  2016-11-13 10:11:52           307         22           0    329           1531.93         164.78            0.00        1696.71\n2       30        1051  2016-10-01 00:48:28           486          0           1    487           2425.14           0.00            8.12        2433.26\n3       33        1051  2016-12-30 08:45:43           495          1           1    497           2470.05           7.49            8.12        2485.66\n4       34        1061  2016-10-19 16:04:11           290         52          23    365           1447.10         389.48          186.76        2023.34\n    ...         ...                  ...           ...        ...         ...    ...               ...            ...             ...            ...\n1693  6596        4121  2016-11-20 13:47:32           172         63         219    454            858.28         471.87         1778.28        3108.43\n1694  6634        4161  2014-06-17 14:02:42            33        305          40    378            164.67        2284.45          324.80        2773.92\n1695  6663        4181  2016-06-24 07:40:53           426         59           0    485           2125.74         441.91            0.00        2567.65\n1696  6855        4451  2014-03-09 07:21:16           241         27          42    310           1202.59         202.23          341.04        1745.86\n1697  6868        4461  2014-08-25 04:07:47            22        185         291    498            109.78        1385.65         2362.92        3858.35\n\n[1698 rows x 11 columns]\n```\n\nPass multiple conditions:\n```python\ndf[(df.total >= 500) & (df.total_amt_usd > 6700)]\n```\n```\n       id  account_id          occurred_at  standard_qty  gloss_qty  poster_qty  total  standard_amt_usd  gloss_amt_usd  poster_amt_usd  total_amt_usd\n0      24        1031  2016-12-25 03:54:27          1148          0         215   1363           5728.52           0.00         1745.80        7474.32\n1     129        1141  2016-12-21 15:52:58           143       1045        2157   3345            713.57        7827.05        17514.84       26055.46\n2     176        1181  2016-04-24 16:47:51          2188         50          12   2250          10918.12         374.50           97.44       11390.06\n3     214        1221  2016-02-04 20:17:54           485       1345          21   1851           2420.15       10074.05          170.52       12664.72\n4     234        1231  2016-11-20 15:16:58           505          0        1355   1860           2519.95           0.00        11002.60       13522.55\n..    ...         ...                  ...           ...        ...         ...    ...               ...            ...             ...            ...\n774  6885        4461  2016-02-04 21:12:41            52        581         276    909            259.48        4351.69         2241.12        6852.29\n775  6890        4461  2016-09-26 08:44:11            42        538         313    893            209.58        4029.62         2541.56        6780.76\n776  6899        4491  2014-03-06 05:22:25           549        523         245   1317           2739.51        3917.27         1989.40        8646.18\n777  6903        4491  2014-09-28 15:53:06            52        601         360   1013            259.48        4501.49         2923.20        7684.17\n778  6906        4491  2015-01-24 07:15:47            54        621         282    957            269.46        4651.29         2289.84        7210.59\n\n[779 rows x 11 columns]\n```\n\nJust like in Pandas you can also save the filtered dataframe/table to a variable\nand perform operations on it \n```python\nfiltered_df = df[(df.total >= 500) & (df.total_amt_usd > 4700)]\n\nfiltered_df.shape\n```\n```\n(1491, 11)\n```\n\n```python\nfiltered_df.account_id.value_counts()\n```\n```\n{3411: 33,\n 2591: 30,\n 4211: 30,\n 1561: 29,\n 2181: 29,\n 4151: 29,\n 1401: 28,\n ...}\n```\n\nSort a table:\n```python\nfiltered_df.sort_values(\'total_amt_usd\', ascending=False)\n```\n```\n        id  account_id          occurred_at  standard_qty  gloss_qty  poster_qty  total  standard_amt_usd  gloss_amt_usd  poster_amt_usd  total_amt_usd\n0     4016        4251  2016-12-26 08:53:24           521         16       28262  28799           2599.79         119.84       229487.44      232207.07\n1     3892        4161  2016-06-24 13:32:55         22591         13           6  22610         112729.09          97.37           48.72      112875.18\n2     3963        4211  2015-03-30 00:05:30           114      14281           0  14395            568.86      106964.69            0.00      107533.55\n3     5791        2861  2014-10-24 12:06:22             0         10       11691  11701              0.00          74.90        94930.92       95005.82\n4     3778        4101  2016-07-17 14:50:43           475          3       11226  11704           2370.25          22.47        91155.12       93547.84\n    ...         ...                  ...           ...        ...         ...    ...               ...            ...             ...            ...\n1486  4721        1491  2015-06-13 21:50:46            46        426         160    632            229.54        3190.74         1299.20        4719.48\n1487  6528        3991  2016-11-17 06:56:25             0         63         523    586              0.00         471.87         4246.76        4718.63\n1488  4719        1491  2015-04-15 14:19:14            54        485         100    639            269.46        3632.65          812.00        4714.11\n1489  5214        2081  2014-06-06 16:55:16            44        600           0    644            219.56        4494.00            0.00        4713.56\n1490  4081        4291  2015-08-09 17:47:52           302        416          11    729           1506.98        3115.84           89.32        4712.14\n\n[1491 rows x 11 columns]\n```\n\nYou can also sort by multiple columns with either a `list` or `dict`\'\n```python\nfiltered_df.sort_values({\'account_id\': \'asc\', \'total_amt_usd\': \'desc\'})\n```\n```\n        id  account_id          occurred_at  standard_qty  gloss_qty  poster_qty  total  standard_amt_usd  gloss_amt_usd  poster_amt_usd  total_amt_usd\n0     4308        1001  2015-12-04 04:01:09           526        597         287   1410           2624.74        4471.53         2330.44        9426.71\n1     4309        1001  2016-01-02 00:59:09           566        645         194   1405           2824.34        4831.05         1575.28        9230.67\n2     4316        1001  2016-08-28 06:50:58           557        572         255   1384           2779.43        4284.28         2070.60        9134.31\n3     4317        1001  2016-09-26 23:22:47           507        614         226   1347           2529.93        4598.86         1835.12        8963.91\n4     4314        1001  2016-05-31 21:09:48           531        603         209   1343           2649.69        4516.47         1697.08        8863.24\n    ...         ...                  ...           ...        ...         ...    ...               ...            ...             ...            ...\n1486  6900        4491  2014-05-05 00:03:19            33        508         283    824            164.67        3804.92         2297.96        6267.55\n1487  6905        4491  2014-12-26 21:39:04           218        366         283    867           1087.82        2741.34         2297.96        6127.12\n1488  6896        4491  2013-12-08 06:34:23            43        520         242    805            214.57        3894.80         1965.04        6074.41\n1489  6901        4491  2014-07-31 05:05:06            12        509         262    783             59.88        3812.41         2127.44        5999.73\n1490  6898        4491  2014-02-04 03:04:08            34        517         205    756            169.66        3872.33         1664.60        5706.59\n\n[1491 rows x 11 columns]\n```\n\n```python\nsorted_df = filtered_df.sort_values({\'account_id\': \'asc\', \'total_amt_usd\': \'desc\'})\nfor row in sorted_df:\n    print(row)\n```\n```\n(4308, 1001, \'2015-12-04 04:01:09\', 526, 597, 287, 1410, 2624.74, 4471.53, 2330.44, 9426.71)\n(4309, 1001, \'2016-01-02 00:59:09\', 566, 645, 194, 1405, 2824.34, 4831.05, 1575.28, 9230.67)\n(4316, 1001, \'2016-08-28 06:50:58\', 557, 572, 255, 1384, 2779.43, 4284.28, 2070.6, 9134.31)\n(4317, 1001, \'2016-09-26 23:22:47\', 507, 614, 226, 1347, 2529.93, 4598.86, 1835.12, 8963.91)\n(4314, 1001, \'2016-05-31 21:09:48\', 531, 603, 209, 1343, 2649.69, 4516.47, 1697.08, 8863.24)\n(4307, 1001, \'2015-11-05 03:25:21\', 506, 612, 203, 1321, 2524.94, 4583.88, 1648.36, 8757.18)\n(4311, 1001, \'2016-03-02 15:40:29\', 498, 605, 204, 1307, 2485.02, 4531.45, 1656.48, 8672.95)\n...\n```\n---\n\n## Cache\n\nWhen initializing the `Database()` object there are a few parameters which will \ndetermine how the output is cached (if it is at all).\n\n\n* `cache` (True/False, default True)\n\nBy default, all the SQL queries are cached, \nso whenever you do `db.table.col.median()` it will calculate the median\nand cache the result for next time, this is done for almost all `Table` and `Column` \nproperties (where the output is always the same or not very large).\n\n* `populate_cache` (True/False, default False)\n\nIf the parameter is set to True it will loop through all the tables and columns \nin the database and populate the cache, so whenever you do `table.shape` or \n`col.avg()` the result will already be there.\nNote that populating the cache can take quite some time for larger databases.\n\n* `max_item_size`(int, default 2), and `max_dict_size` (int, default 100)\n\nThe two parameters determine the maximum size in Megabytes for the element to be cached.\nFor example: if the output of `col.value_counts()` was 2.1MB and max_item_size is set to 2MB,\nthen the output would not be cached since it goes over the limit.\n\nAnd if the output was 1.9MB and the current cache size was 99MB then it would not\ncache the output of the function because otherwise the whole cache-dict size would\nbe above 100MB.\n\n[//]: # (One caveat is that when the output of the query is very large, for example:)\n\n[//]: # (if you do `db.table.col.value_counts&#40;&#41;` and the column values are unique, then)\n\n[//]: # (`len&#40;output&#41;` will be equal to `len&#40;col&#41;`)\n\n---\nAnd finally, you can pass an SQL query to `db.query()` which will return a Pandas `DataFrame` with the results:\n```python\nquery = """\nSELECT * FROM accounts\nJOIN sales_reps\n    ON sales_reps.id = accounts.sales_rep_id\nJOIN region\n    ON region.id = sales_reps.region_id\n"""\ndb.query(query)\n```\n```\n           long  ...    id_2               name_2  region_id id_3     name_3  \n0    -75.103297  ...  321500        Samuel Racine          1    1  Northeast  \n1    -73.849374  ...  321510         Eugena Esser          1    1  Northeast  \n2    -76.084009  ...  321520      Michel Averette          1    1  Northeast  \n3    -75.763898  ...  321530        Renetta Carew          1    1  Northeast  \n4    -75.284998  ...  321540          Cara Clarke          1    1  Northeast  \n..          ...  ...     ...                  ...        ...  ...        ...  \n346 -122.655247  ...  321970  Georgianna Chisholm          4    4       West  \n347 -122.681500  ...  321960  Maryanna Fiorentino          4    4       West  \n348 -122.669460  ...  321970  Georgianna Chisholm          4    4       West  \n349 -122.671880  ...  321960  Maryanna Fiorentino          4    4       West  \n350 -122.657145  ...  321970  Georgianna Chisholm          4    4       West  \n\n[351 rows x 12 columns]\n```\n\nClose the connection\n```python\ndb.exit()\n```\n\n ---\n\nTODO:\n - [x] Move Table and Column objects to a dictionary\n - [x] Support mathematical operations between Column objects (db.table.col1 * db.table.col2)\n - [x] Replace `to_string()` in `__repr__()` with a custom one\n - [x] Add filter function\n - [x] Add cache system for most common properties\n - [ ] Add Group-By method\n',
    'author': 'Shneor Elmaleh',
    'author_email': '770elmo@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/shner-elmo/db-wrapper',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)



from .connection import Database
from .utils import concat
from .expression import Expression  # to pass a custom Expression object in Table.filter or Column.filter

__all__ = ['Database', 'concat', 'Expression']



from __future__ import annotations

import sqlite3
from typing import TYPE_CHECKING

from .utils import get_mb_size
if TYPE_CHECKING:
    from .table import Table


class CacheDict(dict):
    def __setitem__(self, key: str, value: list[tuple]) -> None:
        """
        Add key and value to cache

        Set the SQL query as the key, and the output as the value.

        :param key: str, SQL query
        :param value: list, list with query output (Cursor.fetchall())
        :raise TypeError: if not isinstance(key, str) or not isinstance(val, str)
        :return: None
        """
        if not isinstance(key, str):
            raise TypeError(f'Key must be of type str not {type(key)}')
        if not isinstance(value, list):
            raise TypeError(f'Value must be of type list not {type(value)}')

        super().__setitem__(key, value)

    def __str__(self) -> str:
        """ Get amount of items in the dictionary """
        return f'Cache items: {len(self)}'

    def __repr__(self) -> str:
        """ Get representation of dictionary """
        return super().__repr__()


class Cache(CacheDict):
    """
    A class for managing the cache for all the SQL queries
    """
    def __init__(self, conn: sqlite3.Connection, cache_output: bool, max_item_size: float = 2.0,
                 max_dict_size: float = 100.0) -> None:
        """
        Initialize the cache

        :param conn: Sqlite3 Connection
        :param cache_output: bool, Cache output of SQL queries ?
        :param max_item_size: float, max cache item size in MB (key + value)
        :param max_dict_size: float, max cache dict size in MB
        """
        super().__init__()
        self.conn = conn
        self.cache_output = cache_output
        self.max_item_size = max_item_size
        self.max_dict_size = max_dict_size

        self.mb_size = 0
        self._ready_count = 0

    @property
    def is_ready(self) -> bool:
        """
        Return true if cache is populated with all tables
        """
        tables = [x[0] for x in self.execute("SELECT name FROM sqlite_master WHERE type='table'")]
        return self._ready_count == len(tables)

    def execute(self, query: str) -> list[tuple]:
        """
        Execute an SQL query and save the result in cache for next time

        :param query: str
        :return: list with query results
        """
        if not self.cache_output:
            with self.conn as cursor:
                return cursor.execute(query).fetchall()

        if query in self:
            return self[query]

        with self.conn as cursor:
            query_out = cursor.execute(query).fetchall()

        out_size = get_mb_size(query, query_out)
        if out_size <= self.max_item_size and out_size + self.mb_size <= self.max_dict_size:
            self[query] = query_out
            self.mb_size += out_size

        return query_out

    def populate_table(self, table: Table) -> None:
        """
        Call the most common methods for each column in the table to start populating the cache-dict

        :param table: Table, table object
        :return: None
        """
        getattr(table, 'len')
        getattr(table, 'columns')

        for _, col in table.items():
            getattr(col, 'type')
            getattr(col, 'sql_type')
            getattr(col, 'len')

            col.count(),
            col.null_count(),
            col.min(),
            col.max(),
            col.describe(),

            if col.data_is_numeric():
                col.sum(),
                col.avg(),
                col.median(),

            if col.type in (str, int) and len(table) < 1_000_000:
                col.mode(),
                col.unique(),
                col.value_counts()

        self._ready_count += 1
